@@ -6,14 +6,13 @@ import copy
 import logging
 import os.path
 import pkgutil
-import settings
 import threading
 
 from collections import defaultdict
-from typing import Any, ClassVar, Dict, List, Set, TextIO
+from settings import FilePath, Group, UserFilePath
+from typing import Any, ClassVar, Dict, List, Set, TextIO, Tuple
 
-from BaseClasses import CollectionState, ItemClassification, MultiWorld, Tutorial, Item
-from Fill import fill_restrictive, FillError
+from BaseClasses import CollectionState, Entrance, Item, ItemClassification, MultiWorld, Region, Tutorial
 from worlds.AutoWorld import WebWorld, World
 from entrance_rando import ERPlacementState
 from .client import PokemonFRLGClient
@@ -21,29 +20,31 @@ from .client import PokemonFRLGClient
 from .data import (data, ability_name_map, ALL_SPECIES, APWORLD_VERSION, LEGENDARY_POKEMON, NAME_TO_SPECIES_ID,
                    POPTRACKER_CHECKSUM, LocationCategory, EventData, EvolutionMethodEnum, FlyData, MapData,
                    MiscPokemonData, MoveData, move_name_map, SpeciesData, StarterData, TrainerData, TradePokemonData)
-from .entrances import shuffle_entrances
+from .entrances import set_hint_entrances, shuffle_entrances
 from .groups import item_groups, location_groups
-from .items import (PokemonFRLGItem, add_starting_items, create_item_name_to_id_map, get_random_item,
-                    get_item_classification)
+from .items import (PokemonFRLGItem, PokemonFRLGGlitchedToken, add_starting_items, create_item_name_to_id_map,
+                    get_random_item, get_item_classification)
 from .level_scaling import level_scaling
 from .locations import (PokemonFRLGLocation, create_location_name_to_id_map, create_locations,
-                        place_unrandomized_items, place_shop_items, set_free_fly)
-from .options import (PokemonFRLGOptions, CardKey, CeruleanCaveRequirement, Dexsanity, DungeonEntranceShuffle,
-                      FishingRods, FlashRequired, FreeFlyLocation, GameVersion, Goal, IslandPasses,
-                      RandomizeLegendaryPokemon, RandomizeMiscPokemon, RandomizeWildPokemon, ShuffleBadges,
-                      ShuffleFlyUnlocks, ShuffleHiddenItems, ShufflePokedex, ShuffleRunningShoes, TownMapFlyLocation,
-                      Trainersanity, ViridianCityRoadblock)
+                        place_unrandomized_items, place_shop_items, set_free_fly, shuffle_badges)
+from .options import (PokemonFRLGOptions, CardKey, CeruleanCaveRequirement, Dexsanity, FishingRods, FlashRequired,
+                      FreeFlyLocation, GameVersion, Goal, IslandPasses, MixEntranceWarpPools, RandomizeLegendaryPokemon,
+                      RandomizeMiscPokemon, RandomizeWildPokemon, ShuffleBadges, ShuffleBuildingEntrances,
+                      ShuffleDungeonEntrances, ShuffleFlyUnlocks, ShuffleHiddenItems, ShufflePokedex,
+                      ShuffleRunningShoes, TownMapFlyLocation, Trainersanity, ViridianCityRoadblock)
 from .pokemon import (add_hm_compatability, randomize_abilities, randomize_base_stats, randomize_damage_categories,
                       randomize_legendaries, randomize_misc_pokemon, randomize_moves, randomize_move_types,
                       randomize_requested_trade_pokemon, randomize_starters, randomize_tm_hm_compatibility,
                       randomize_tm_moves, randomize_trainer_parties, randomize_types, randomize_wild_encounters)
-from .regions import starting_town_map, create_indirect_conditions, create_regions
+from .regions import starting_town_map, create_indirect_conditions, create_regions, PokemonFRLGRegion
 from .rules import PokemonFRLGLogic, set_hm_compatible_pokemon, set_logic_options, set_rules, verify_hm_accessibility
 from .rom import PokemonFRLGPatchData, PokemonFireRedProcedurePatch, PokemonLeafGreenProcedurePatch, write_tokens
 from .sanity_check import validate_regions
+from .universal_tracker import (POPTRACKER_LOCATIONS, map_page_index, ut_reconnect_found_entrances, ut_set_locations,
+                                ut_set_maps, ut_set_options)
 from .util import int_to_bool_array, HM_TO_COMPATIBILITY_ID
 
-# Try adding the Pokemon Gen 3 Adjuster
+# Try adding the Pokémon Gen 3 Adjuster
 try:
     from worlds._pokemon_gen3_adjuster import __init__
 except ImportError:
@@ -74,21 +75,26 @@ class PokemonFRLGWebWorld(WebWorld):
     tutorials = [setup_en, adjuster_en]
 
 
-class PokemonFRLGSettings(settings.Group):
-    class PokemonFireRedRomFile(settings.UserFilePath):
+class PokemonFRLGSettings(Group):
+    class PokemonFireRedRomFile(UserFilePath):
         """File name of your English Pokémon FireRed ROM"""
         description = "Pokemon FireRed ROM File"
         copy_to = "Pokemon - FireRed Version (USA, Europe).gba"
         md5s = PokemonFireRedProcedurePatch.hash
 
-    class PokemonLeafGreenRomFile(settings.UserFilePath):
+    class PokemonLeafGreenRomFile(UserFilePath):
         """File name of your English Pokémon LeafGreen ROM"""
         description = "Pokemon LeafGreen ROM File"
         copy_to = "Pokemon - LeafGreen Version (USA, Europe).gba"
         md5s = PokemonLeafGreenProcedurePatch.hash
 
+    class UTPoptrackerPath(FilePath):
+        description = "Pokemon FRLG Poptracker Pack Zip File"
+        required = False
+
     firered_rom_file: PokemonFireRedRomFile = PokemonFireRedRomFile(PokemonFireRedRomFile.copy_to)
     leafgreen_rom_file: PokemonLeafGreenRomFile = PokemonLeafGreenRomFile(PokemonLeafGreenRomFile.copy_to)
+    ut_poptracker_path: UTPoptrackerPath | str = UTPoptrackerPath()
 
 
 class PokemonFRLGWorld(World):
@@ -120,9 +126,21 @@ class PokemonFRLGWorld(World):
     required_client_version = (0, 6, 0)
     origin_region_name = "Title Screen"
 
+    ut_can_gen_without_yaml = True
+    glitches_item_name = PokemonFRLGGlitchedToken.TOKEN_NAME
+    is_universal_tracker: bool
+    found_entrances_datastorage_key: List[str] = []
+    tracker_world = {
+        "external_pack_key": "ut_poptracker_path",
+        "poptracker_name_mapping": {k: data.locations[v].flag for k, v in POPTRACKER_LOCATIONS.items()},
+        "map_page_index": map_page_index,
+        "map_page_setting_key": "pokemon_frlg_map_{team}_{player}",
+    }
+
     logic: PokemonFRLGLogic
     patch_data: PokemonFRLGPatchData
     starting_town: str
+    starting_respawn: str
     free_fly_location_id: int
     town_map_fly_location_id: int
     modified_species: Dict[int, SpeciesData]
@@ -155,16 +173,18 @@ class PokemonFRLGWorld(World):
     pre_fill_items: List[PokemonFRLGItem]
     fly_destination_data: Dict[str, FlyData]
     er_placement_state: ERPlacementState | None
-    er_spoiler_names: List[str]
+    er_entrances: List[Tuple[Entrance, Region]]
     moves_by_type: Dict[int, Set[int]]
     cerulean_cave_included: bool
     auth: bytes
 
     def __init__(self, multiworld, player):
         super(PokemonFRLGWorld, self).__init__(multiworld, player)
+        self.is_universal_tracker = hasattr(self.multiworld, "generation_is_fake")
         self.logic = PokemonFRLGLogic(player, self.item_id_to_name)
         self.patch_data = PokemonFRLGPatchData()
         self.starting_town = "SPAWN_PALLET_TOWN"
+        self.starting_respawn = "SPAWN_PALLET_TOWN"
         self.free_fly_location_id = 0
         self.town_map_fly_location_id = 0
         self.modified_species = copy.deepcopy(data.species)
@@ -189,7 +209,7 @@ class PokemonFRLGWorld(World):
         self.pre_fill_items = []
         self.fly_destination_data = {}
         self.er_placement_state = None
-        self.er_spoiler_names = []
+        self.er_entrances = []
         self.moves_by_type = {}
         self.cerulean_cave_included = True
         self.finished_level_scaling = threading.Event()
@@ -202,6 +222,11 @@ class PokemonFRLGWorld(World):
         return get_random_item(self, ItemClassification.filler)
 
     def generate_early(self) -> None:
+        if self.is_universal_tracker:
+            ut_set_options(self)
+            ut_set_maps(self)
+            ut_set_locations(self)
+
         self.blacklisted_wild_pokemon = {
             species.species_id for species in self.modified_species.values()
             if species.name in self.options.wild_pokemon_blacklist.value
@@ -241,6 +266,10 @@ class PokemonFRLGWorld(World):
         self.blacklisted_moves = {move_name_map[name] for name in self.options.move_blacklist.value}
         self.blacklisted_tm_tutor_moves = {move_name_map[name] for name in self.options.tm_tutor_moves_blacklist.value}
 
+        if "All" in self.options.mix_entrance_warp_pools.value:
+            for key in MixEntranceWarpPools.valid_keys:
+                self.options.mix_entrance_warp_pools.value.add(key)
+
         # Modify options that are incompatible with each other
         if self.options.kanto_only:
             if self.options.goal == Goal.option_champion_rematch:
@@ -254,6 +283,18 @@ class PokemonFRLGWorld(World):
                                 self.player, self.player_name)
                 self.options.cerulean_cave_requirement.value = CeruleanCaveRequirement.option_champion
 
+        if self.options.decouple_entrances_warps:
+            if self.options.shuffle_buildings == ShuffleBuildingEntrances.option_simple:
+                logging.warning("Pokemon FRLG: Simple Building Shuffle for player %s (%s) is "
+                                "incompatible with Decoupled Entrances. Setting shuffle to Restricted.",
+                                self.player, self.player_name)
+                self.options.shuffle_buildings.value = ShuffleBuildingEntrances.option_restricted
+            if self.options.shuffle_dungeons == ShuffleDungeonEntrances.option_simple:
+                logging.warning("Pokemon FRLG: Simple Dungeon Shuffle for player %s (%s) is "
+                                "incompatible with Decoupled Entrances. Setting shuffle to Restricted.",
+                                self.player, self.player_name)
+                self.options.shuffle_dungeons.value = ShuffleDungeonEntrances.option_restricted
+
         # Check if Ceruelan Cave should be included in this world
         if (not self.options.post_goal_locations and
                 self.options.goal == Goal.option_champion and
@@ -364,8 +405,8 @@ class PokemonFRLGWorld(World):
         if self.options.shuffle_tm_case:
             self.itempool.append(self.create_item("TM Case"))
             self.itempool.remove(filler_items.pop())
-        if self.options.shuffle_ledge_jump:
-            self.itempool.append(self.create_item("Ledge Jump"))
+        if self.options.shuffle_jumping_shoes:
+            self.itempool.append(self.create_item("Jumping Shoes"))
             self.itempool.remove(filler_items.pop())
 
         # Add key items that are relevant in Kanto Only to the itempool
@@ -377,8 +418,7 @@ class PokemonFRLGWorld(World):
 
         # Remove copies unique items based on how many are in the start inventory
         unique_items: Set[str] = set(item_groups["Unique Items"] |
-                                     item_groups["Progressive Items"] |
-                                     item_groups["Abilities"])
+                                     item_groups["Progressive Items"])
         for item in self.multiworld.precollected_items[self.player]:
             assert isinstance(item, PokemonFRLGItem)
             if item.name in unique_items:
@@ -397,7 +437,7 @@ class PokemonFRLGWorld(World):
             if not location.can_reach(state):
                 evolution_region.locations.remove(location)
 
-        if self.options.dexsanity != Dexsanity.special_range_names["none"]:
+        if self.options.dexsanity != Dexsanity.special_range_names["none"] and not self.is_universal_tracker:
             # Delete dexsanity locations that are not in logic in an all state since they aren't accessible
             pokedex_region = self.multiworld.get_region("Pokedex", self.player)
             for location in pokedex_region.locations.copy():
@@ -428,42 +468,10 @@ class PokemonFRLGWorld(World):
     def connect_entrances(self) -> None:
         set_free_fly(self)
         if not self.options.shuffle_badges:
-            self.shuffle_badges()
-        if self.options.dungeon_entrance_shuffle != DungeonEntranceShuffle.option_off:
-            shuffle_entrances(self)
+            shuffle_badges(self)
+            verify_hm_accessibility(self)
+        if shuffle_entrances(self):
             verify_hm_accessibility(self)
-
-    def shuffle_badges(self) -> None:
-        badge_items = []
-        badge_items.extend(self.get_pre_fill_items())
-        self.pre_fill_items.clear()
-        locations: List[PokemonFRLGLocation] = self.get_locations()
-        for attempt in range(5):
-            badge_locations: List[PokemonFRLGLocation] = [
-                loc for loc in locations if loc.name in location_groups["Gym Prizes"] and loc.item is None
-            ]
-            state = self.get_world_collection_state()
-            # Try to place badges with current Pokemon and HM access
-            # If it can't, try with guaranteed HM access and fix it later
-            if attempt > 1:
-                self.logic.guaranteed_hm_access = True
-            state.sweep_for_advancements()
-            self.random.shuffle(badge_items)
-            self.random.shuffle(badge_locations)
-            fill_restrictive(self.multiworld, state, badge_locations.copy(), badge_items,
-                             single_player_placement=True, lock=True, allow_partial=True, allow_excluded=True)
-            if len(badge_items) > 8 - len(badge_locations):
-                for location in badge_locations:
-                    if location.item:
-                        badge_items.append(location.item)
-                        location.item = None
-                continue
-            else:
-                break
-        else:
-            raise FillError(f"Failed to place badges for player {self.player}")
-        self.logic.guaranteed_hm_access = False
-        verify_hm_accessibility(self)
 
     def generate_basic(self) -> None:
         # Create auth
@@ -532,18 +540,6 @@ class PokemonFRLGWorld(World):
         out_file_name = self.multiworld.get_out_file_name_base(self.player)
         patch.write(os.path.join(output_directory, f"{out_file_name}{patch.patch_file_ending}"))
 
-        del self.modified_species
-        del self.modified_maps
-        del self.modified_starters
-        del self.modified_events
-        del self.modified_legendary_pokemon
-        del self.modified_misc_pokemon
-        del self.modified_trade_pokemon
-        del self.modified_trainers
-        del self.modified_tmhm_moves
-        del self.modified_moves
-        del self.modified_type_damage_categories
-
     def write_spoiler_header(self, spoiler_handle: TextIO) -> None:
         if self.options.random_starting_town:
             starting_town = starting_town_map[self.starting_town]
@@ -558,12 +554,12 @@ class PokemonFRLGWorld(World):
             spoiler_handle.write(f"Town Map Fly Location:           {town_map_fly_location.item.name}\n")
 
     def write_spoiler(self, spoiler_handle: TextIO) -> None:
-        # Add dungeon entrances to the spoiler log if they are shuffled
-        if self.options.dungeon_entrance_shuffle != DungeonEntranceShuffle.option_off:
-            spoiler_handle.write(f"\n\nDungeon Entrances ({self.multiworld.player_name[self.player]}):\n\n")
-            for entrance, exit in sorted(self.er_placement_state.pairings):
-                if entrance in self.er_spoiler_names:
-                    spoiler_handle.write(f"{entrance} <=> {exit}\n")
+        # Add entrances to the spoiler log if they are shuffled
+        if self.er_placement_state:
+            spoiler_handle.write(f"\n\nEntrances ({self.multiworld.player_name[self.player]}):\n\n")
+            for entrance_name, exit_name in sorted(self.er_placement_state.pairings):
+                entrance = self.get_entrance(entrance_name)
+                spoiler_handle.write(f"{entrance_name} => {entrance.connected_region}\n")
 
         # Add fly destinations to the spoiler log if they are randomized
         if self.options.randomize_fly_destinations:
@@ -599,6 +595,7 @@ class PokemonFRLGWorld(World):
                 spoiler_handle.write(line)
 
     def extend_hint_information(self, hint_data):
+        hint_data[self.player] = {}
         if self.options.dexsanity != Dexsanity.special_range_names["none"]:
             species_locations = defaultdict(set)
 
@@ -610,10 +607,17 @@ class PokemonFRLGWorld(World):
                     pokemon_name = location.item.name.replace("Static ", "")
                     species_locations[pokemon_name].add(location.spoiler_name)
 
-            hint_data[self.player] = {
-                self.location_name_to_id[f"Pokedex - {species}"]: ", ".join(sorted(maps))
-                for species, maps in species_locations.items()
-            }
+            for species, maps in species_locations.items():
+                hint_data[self.player][self.location_name_to_id[f"Pokedex - {species}"]] = ", ".join(sorted(maps))
+
+        if self.er_placement_state is not None:
+            set_hint_entrances(self)
+            for region in self.get_regions():
+                assert isinstance(region, PokemonFRLGRegion)
+                if region.entrance_hints:
+                    for location in region.locations:
+                        if not location.is_event:
+                            hint_data[self.player][location.address] = ", ".join(sorted(region.entrance_hints))
 
     def modify_multidata(self, multidata: Dict[str, Any]):
         multidata["connect_names"][base64.b64encode(self.auth).decode("ascii")] = \
@@ -625,24 +629,52 @@ class PokemonFRLGWorld(World):
             "goal",
             "skip_elite_four",
             "kanto_only",
+            "random_starting_town",
+            "shuffle_pokemon_centers",
+            "shuffle_gyms",
+            "shuffle_marts",
+            "shuffle_harbors",
+            "shuffle_buildings",
+            "shuffle_dungeons",
+            "shuffle_interiors",
+            "shuffle_warp_tiles",
+            "shuffle_dropdowns",
+            "decouple_entrances_warps",
+            "randomize_fly_destinations",
             "shuffle_badges",
             "shuffle_hidden",
             "extra_key_items",
             "shopsanity",
+            "vending_machines",
+            "prizesanity",
+            "shop_slots",
             "rematchsanity",
+            "rematch_requirements",
             "famesanity",
             "shuffle_fly_unlocks",
             "pokemon_request_locations",
+            "shuffle_pokedex",
+            "shuffle_running_shoes",
+            "shuffle_berry_pouch",
+            "shuffle_tm_case",
+            "shuffle_jumping_shoes",
+            "post_goal_locations",
             "card_key",
             "island_passes",
+            "fishing_rods",
             "split_teas",
             "gym_keys",
-            "post_goal_locations",
             "itemfinder_required",
             "flash_required",
             "fame_checker_required",
-            "bicycle_requires_ledge_jump",
+            "bicycle_requires_jumping_shoes",
             "acrobatic_bicycle",
+            "evolutions_required",
+            "evolution_methods_required",
+            "viridian_city_roadblock",
+            "pewter_city_roadblock",
+            "modify_world_state",
+            "additional_dark_caves",
             "remove_badge_requirement",
             "oaks_aide_route_2",
             "oaks_aide_route_10",
@@ -650,10 +682,6 @@ class PokemonFRLGWorld(World):
             "oaks_aide_route_16",
             "oaks_aide_route_15",
             "fossil_count",
-            "viridian_city_roadblock",
-            "pewter_city_roadblock",
-            "modify_world_state",
-            "additional_dark_caves",
             "viridian_gym_requirement",
             "viridian_gym_count",
             "route22_gate_requirement",
@@ -665,40 +693,62 @@ class PokemonFRLGWorld(World):
             "elite_four_rematch_count",
             "cerulean_cave_requirement",
             "cerulean_cave_count",
+            "free_fly_location",
+            "town_map_fly_location",
             "provide_hints",
             "death_link"
         )
+
+        game_version = self.options.game_version.current_key
+
         slot_data["trainersanity"] = 1 if self.options.trainersanity != Trainersanity.special_range_names["none"] else 0
+        slot_data["trainersanity_locations"] = [loc.address for loc in self.get_locations()
+                                                if loc.category == LocationCategory.TRAINER
+                                                or loc.category == LocationCategory.TRAINER_REMATCH]
+        slot_data["dexsanity"] = 1 if self.options.dexsanity != Dexsanity.special_range_names["none"] else 0
+        slot_data["dexsanity_locations"] = [loc.address for loc in self.get_locations()
+                                            if loc.category == LocationCategory.POKEDEX]
         slot_data["elite_four_rematch_requirement"] = self.options.elite_four_requirement.value
         slot_data["starting_town"] = data.constants[self.starting_town]
         slot_data["free_fly_location_id"] = self.free_fly_location_id
         slot_data["town_map_fly_location_id"] = self.town_map_fly_location_id
+
         if self.options.randomize_fly_destinations:
-            slot_data["randomize_fly_destinations"] = {}
+            slot_data["fly_destinations"] = {}
             for exit in self.get_region("Sky").exits:
-                slot_data["randomize_fly_destinations"][exit.name] = exit.connected_region.name
-        if self.options.dungeon_entrance_shuffle != DungeonEntranceShuffle.option_off:
-            slot_data["dungeon_entrance_shuffle"] = []
+                slot_data["fly_destinations"][exit.name] = exit.connected_region.name
+
+        if self.er_placement_state is not None:
+            slot_data["entrances"] = {}
             for source, dest in self.er_placement_state.pairings:
-                slot_data["dungeon_entrance_shuffle"].append(source)
+                slot_data["entrances"][source] = self.get_entrance(source).connected_region.name
+
         slot_data["wild_encounters"] = {}
         slot_data["static_encounters"] = {}
-        for location in self.get_locations():
-            assert isinstance(location, PokemonFRLGLocation)
-            if location.category == LocationCategory.EVENT_WILD_POKEMON:
-                national_dex_id = data.species[NAME_TO_SPECIES_ID[location.item.name]].national_dex_number
-                if national_dex_id not in slot_data["wild_encounters"]:
-                    slot_data["wild_encounters"][national_dex_id] = []
-                slot_data["wild_encounters"][national_dex_id].append(location.name)
-            elif location.category in (LocationCategory.EVENT_STATIC_POKEMON,
-                                       LocationCategory.EVENT_LEGENDARY_POKEMON):
-                pokemon_name = location.item.name.replace("Static ", "")
-                national_dex_id = data.species[NAME_TO_SPECIES_ID[pokemon_name]].national_dex_number
-                slot_data["static_encounters"][location.name] = national_dex_id
+        wild_locations = [loc for loc in self.get_locations() if loc.category == LocationCategory.EVENT_WILD_POKEMON]
+        static_locations = [loc for loc in self.get_locations()
+                            if loc.category == LocationCategory.EVENT_STATIC_POKEMON
+                            or loc.category == LocationCategory.EVENT_LEGENDARY_POKEMON]
+        for location in wild_locations:
+            national_dex_id = data.species[NAME_TO_SPECIES_ID[location.item.name]].national_dex_number
+            if location.encounter_key not in slot_data["wild_encounters"]:
+                slot_data["wild_encounters"][location.encounter_key] = set()
+            slot_data["wild_encounters"][location.encounter_key].add(national_dex_id)
+        for location in static_locations:
+            pokemon_name = location.item.name.replace("Static ", "")
+            national_dex_id = data.species[NAME_TO_SPECIES_ID[pokemon_name]].national_dex_number
+            slot_data["static_encounters"][location.encounter_key] = national_dex_id
+
+        slot_data["tm_hm_compatibility"] = {v.name: v.tm_hm_compatibility for v in self.modified_species.values()}
+        slot_data["requested_trade_pokemon"] = {k: v.requested_species_id[game_version]
+                                                for k, v in self.modified_trade_pokemon.items()}
+        slot_data["resort_gorgeous_pokemon"] = self.logic.resort_gorgeous_pokemon
         slot_data["poptracker_checksum"] = POPTRACKER_CHECKSUM
         return slot_data
 
     def create_item(self, name: str) -> "PokemonFRLGItem":
+        if name == self.glitches_item_name:
+            return PokemonFRLGGlitchedToken(self.player)
         return self.create_item_by_id(self.item_name_to_id[name])
 
     def create_item_by_id(self, item_id: int):
@@ -749,3 +799,19 @@ class PokemonFRLGWorld(World):
             return True
         else:
             return False
+
+    # Universal Tracker
+    @property
+    def ut_slot_data(self) -> dict[str, Any]:
+        if hasattr(self.multiworld, "re_gen_passthrough"):
+            return self.multiworld.re_gen_passthrough[self.game]
+        else:
+            return {}
+
+    @staticmethod
+    def interpret_slot_data(slot_data: dict[str, Any]):
+        return slot_data
+
+    def reconnect_found_entrances(self, found_key: str, data_storage_value: Any) -> None:
+        if data_storage_value:
+            ut_reconnect_found_entrances(self, found_key)
