@@ -1,5 +1,6 @@
 from typing import TYPE_CHECKING, Dict, List, Set
 from BaseClasses import CollectionState, Location, LocationProgressType, Region, ItemClassification
+from Fill import FillError, fill_restrictive
 from .data import data, LocationCategory, fly_blacklist_map, TRAINER_REMATCH_MAP
 from .groups import location_groups
 from .items import PokemonFRLGItem, get_random_item, update_renewable_to_progression
@@ -65,6 +66,8 @@ class PokemonFRLGLocation(Location):
     location_id: str | None
     scaling_ids: List[str] | None
     spoiler_name: str
+    encounter_key: str
+    hint_entrances: List[str]
 
     def __init__(
             self,
@@ -77,7 +80,8 @@ class PokemonFRLGLocation(Location):
             default_item_id: int | None = None,
             location_id: str | None = None,
             scaling_ids: List[str] | None = None,
-            spoiler_name: str | None = None) -> None:
+            spoiler_name: str | None = None,
+            encounter_key: str | None = None) -> None:
         super().__init__(player, name, address, parent)
         self.default_item_id = default_item_id
         self.item_address = item_address
@@ -85,6 +89,7 @@ class PokemonFRLGLocation(Location):
         self.location_id = location_id
         self.scaling_ids = scaling_ids
         self.spoiler_name = spoiler_name if spoiler_name is not None else name
+        self.encounter_key = encounter_key
 
 
 def create_location_name_to_id_map() -> Dict[str, int]:
@@ -166,6 +171,8 @@ def create_locations(world: "PokemonFRLGWorld", regions: Dict[str, Region]) -> N
         included_types.add("Extra Key Items")
     if world.options.trainersanity != Trainersanity.special_range_names["none"]:
         included_types.add("Trainersanity")
+        if world.options.rematchsanity:
+            included_types.add("Rematchsanity")
     if world.options.dexsanity != Dexsanity.special_range_names["none"]:
         included_types.add("Dexsanity")
     if world.options.famesanity:
@@ -192,14 +199,24 @@ def create_locations(world: "PokemonFRLGWorld", regions: Dict[str, Region]) -> N
         for location_id in included_locations:
             if exclude_location(location_id):
                 continue
+            if world.is_universal_tracker:
+                location = data.locations[location_id]
+                if ((location.category == LocationCategory.TRAINER
+                        or location.category == LocationCategory.TRAINER_REMATCH)
+                        and location.flag not in world.ut_slot_data["trainersanity_locations"]):
+                    continue
+                elif (location.category == LocationCategory.POKEDEX
+                      and location.flag not in world.ut_slot_data["dexsanity_locations"]):
+                    continue
             region.locations.append(create_location(location_id))
 
     # Remove trainersanity locations if there are more than the amount specified in the settings
-    if world.options.trainersanity != Trainersanity.special_range_names["none"]:
+    if world.options.trainersanity != Trainersanity.special_range_names["none"] and not world.is_universal_tracker:
         locations: List[PokemonFRLGLocation] = world.get_locations()
         trainer_locations = [loc for loc in locations if loc.category == LocationCategory.TRAINER]
         locs_to_remove = len(trainer_locations) - world.options.trainersanity.value
         if locs_to_remove > 0:
+            rematchsanity = world.options.rematchsanity
             priority_trainer_locations = [loc for loc in trainer_locations
                                           if loc.name in world.options.priority_locations.value]
             non_priority_trainer_locations = [loc for loc in trainer_locations
@@ -210,24 +227,17 @@ def create_locations(world: "PokemonFRLGWorld", regions: Dict[str, Region]) -> N
             for location in trainer_locations:
                 region = location.parent_region
                 region.locations.remove(location)
+                if rematchsanity and location.location_id in TRAINER_REMATCH_MAP:
+                    for location_id in TRAINER_REMATCH_MAP[location.location_id]:
+                        region.locations.remove(world.get_location(data.locations[location_id].name))
                 locs_to_remove -= 1
                 if locs_to_remove <= 0:
                     break
 
-    # Add rematchsanity locations
-    if world.options.rematchsanity:
-        locations: List[PokemonFRLGLocation] = world.get_locations()
-        trainer_locations = [loc for loc in locations if loc.category == LocationCategory.TRAINER]
-        for location in trainer_locations:
-            if location.location_id in TRAINER_REMATCH_MAP:
-                for location_id in TRAINER_REMATCH_MAP[location.location_id]:
-                    location.parent_region.locations.append(create_location(location_id))
-
 
 def place_unrandomized_items(world: "PokemonFRLGWorld") -> None:
     def fill_unrandomized_location(location: Location,
                                             as_event: bool) -> None:
-        assert isinstance(location, PokemonFRLGLocation)
         item = world.create_item_by_id(location.default_item_id)
         if as_event:
             item.code = None
@@ -244,13 +254,21 @@ def place_unrandomized_items(world: "PokemonFRLGWorld") -> None:
     elif world.options.shuffle_fly_unlocks == ShuffleFlyUnlocks.option_exclude_indigo:
         fill_unrandomized_location(world.get_location("Indigo Plateau - Unlock Fly Destination"), False)
 
+    shop_locations = []
     if world.options.shopsanity:
-        shop_locations = [loc for loc in world.get_locations() if "Held Shop Item" in loc.name]
+        shop_locations.extend([loc for loc in world.get_locations() if "Held Shop Item" in loc.name])
+        if not world.options.post_goal_locations and world.options.goal == Goal.option_champion:
+            shop_locations.extend([loc for loc in world.get_locations() if
+                                   loc.name in location_groups["Market Stall"] and
+                                   int(loc.name[-1]) in (3, 4, 5, 8, 9)])
     else:
-        shop_locations = [loc for loc in world.get_locations() if loc.name in location_groups["Shops"]]
+        shop_locations.extend([loc for loc in world.get_locations() if loc.name in location_groups["Shops"]])
+    if not world.options.vending_machines:
+        shop_locations.extend([loc for loc in world.get_locations() if loc.name in location_groups["Vending Machines"]])
+    if not world.options.prizesanity:
+        shop_locations.extend([loc for loc in world.get_locations() if loc.name in location_groups["Prizes"]])
     for location in shop_locations:
         fill_unrandomized_location(location, True)
-        assert isinstance(location.item, PokemonFRLGItem)
         update_renewable_to_progression(location.item)
 
     if world.options.shuffle_pokedex == ShufflePokedex.option_vanilla:
@@ -261,26 +279,40 @@ def place_unrandomized_items(world: "PokemonFRLGWorld") -> None:
 
 
 def place_shop_items(world: "PokemonFRLGWorld") -> None:
-    if not world.options.shopsanity:
+    if (not world.options.shopsanity and not world.options.vending_machines) or world.is_universal_tracker:
         return
 
-    shop_locations = [loc for loc in world.get_locations() if loc.name in location_groups["Shops"] and
-                      loc.item is None]
+    shop_locations = [loc for loc in world.get_locations() if
+                      (loc.name in location_groups["Shops"] or
+                       loc.name in location_groups["Vending Machines"] or
+                       loc.name in location_groups["Prizes"]) and
+                      not loc.is_event]
     shop_items = [world.create_item_by_id(loc.default_item_id) for loc in shop_locations]
     non_progression_shop_locations = [loc for loc in shop_locations if
-                                      "Two Island Town - Market Stall Item" not in loc.name and
+                                      loc.name not in location_groups["Market Stall"] and
+                                      loc.name not in location_groups["Vending Machines"] and
+                                      loc.name not in location_groups["Prizes"] and
                                       int(loc.name[-1]) > world.options.shop_slots.value]
     world.random.shuffle(shop_items)
 
     if not world.options.kanto_only:
         two_island_shop_location_ids = ["SHOP_TWO_ISLAND_1", "SHOP_TWO_ISLAND_7", "SHOP_TWO_ISLAND_2",
-                                        "SHOP_TWO_ISLAND_6",
-                                        "SHOP_TWO_ISLAND_5", "SHOP_TWO_ISLAND_8", "SHOP_TWO_ISLAND_3",
-                                        "SHOP_TWO_ISLAND_4",
-                                        "SHOP_TWO_ISLAND_9"]
+                                        "SHOP_TWO_ISLAND_6", "SHOP_TWO_ISLAND_5", "SHOP_TWO_ISLAND_8",
+                                        "SHOP_TWO_ISLAND_3", "SHOP_TWO_ISLAND_4", "SHOP_TWO_ISLAND_9"]
         for index, location_id in enumerate(two_island_shop_location_ids):
-            if index >= world.options.shop_slots.value:
+            if (index >= world.options.shop_slots.value and
+                    not world.get_location(data.locations[location_id].name).is_event):
                 non_progression_shop_locations.append(world.get_location(data.locations[location_id].name))
+
+
+    renewable_items = []
+    if world.options.vending_machines:
+        renewable_items.extend(["Fresh Water", "Soda Pop", "Lemonade"])
+    elif (world.options.shopsanity and
+          not world.options.kanto_only and
+          (world.options.post_goal_locations or world.options.goal != Goal.option_champion)):
+        renewable_items.append("Lemonade")
+    if renewable_items:
         renewable_locations = [loc for loc in shop_locations if
                                loc.progress_type != LocationProgressType.EXCLUDED and
                                loc not in non_progression_shop_locations]
@@ -289,15 +321,17 @@ def place_shop_items(world: "PokemonFRLGWorld") -> None:
                                    loc not in non_progression_shop_locations]
         if len(renewable_locations) == 0:
             renewable_locations = [loc for loc in shop_locations]
-        renewable_location = world.random.choice(renewable_locations)
-        item = world.create_item("Lemonade")
-        item.classification = ItemClassification.progression
-        renewable_location.place_locked_item(item)
-        world.itempool.remove(item)
-        shop_items.remove(item)
+        world.random.shuffle(renewable_locations)
+        for item_name in renewable_items:
+            renewable_location = renewable_locations.pop()
+            item = world.create_item(item_name)
+            item.classification = ItemClassification.progression
+            renewable_location.place_locked_item(item)
+            world.itempool.remove(item)
+            shop_items.remove(item)
 
     for location in non_progression_shop_locations:
-        if location.item is not None:
+        if location.locked:
             continue
         item = shop_items.pop()
         location.place_locked_item(item)
@@ -352,7 +386,11 @@ def set_free_fly(world: "PokemonFRLGWorld") -> None:
         town_map_fly_list = [fly for fly in fly_item_map.values() if fly not in forbidden_fly_list]
 
     if world.options.free_fly_location:
-        free_fly_location_id = world.random.choice(free_fly_list)
+        if not world.is_universal_tracker:
+            free_fly_location_id = world.random.choice(free_fly_list)
+        else:
+            fly_map_lookup = {v: k for k, v in fly_item_id_map.items()}
+            free_fly_location_id = fly_map_lookup[world.ut_slot_data["free_fly_location_id"]]
         world.free_fly_location_id = fly_item_id_map[free_fly_location_id]
 
         if free_fly_location_id in town_map_fly_list and len(town_map_fly_list) > 1:
@@ -377,7 +415,11 @@ def set_free_fly(world: "PokemonFRLGWorld") -> None:
         start_region.locations.append(free_fly_location)
 
     if world.options.town_map_fly_location:
-        town_map_fly_location_id = world.random.choice(town_map_fly_list)
+        if not world.is_universal_tracker:
+            town_map_fly_location_id = world.random.choice(town_map_fly_list)
+        else:
+            fly_map_lookup = {v: k for k, v in fly_item_id_map.items()}
+            town_map_fly_location_id = fly_map_lookup[world.ut_slot_data["town_map_fly_location_id"]]
         world.town_map_fly_location_id = fly_item_id_map[town_map_fly_location_id]
 
         start_region = world.multiworld.get_region("Title Screen", world.player)
@@ -398,3 +440,37 @@ def set_free_fly(world: "PokemonFRLGWorld") -> None:
         town_map_fly_location.access_rule = lambda state: state.has("Town Map", world.player)
         town_map_fly_location.show_in_spoiler = False
         start_region.locations.append(town_map_fly_location)
+
+def shuffle_badges(world: "PokemonFRLGWorld") -> None:
+    if world.is_universal_tracker:
+        return
+
+    badge_items = []
+    badge_items.extend(world.get_pre_fill_items())
+    world.pre_fill_items.clear()
+    locations: List[PokemonFRLGLocation] = world.get_locations()
+    for attempt in range(5):
+        badge_locations: List[PokemonFRLGLocation] = [
+            loc for loc in locations if loc.name in location_groups["Gym Prizes"] and loc.item is None
+        ]
+        state = world.get_world_collection_state()
+        # Try to place badges with current Pokemon and HM access
+        # If it can't, try with guaranteed HM access and fix it later
+        if attempt > 1:
+            world.logic.guaranteed_hm_access = True
+        state.sweep_for_advancements()
+        world.random.shuffle(badge_items)
+        world.random.shuffle(badge_locations)
+        fill_restrictive(world.multiworld, state, badge_locations.copy(), badge_items,
+                         single_player_placement=True, lock=True, allow_partial=True, allow_excluded=True)
+        if len(badge_items) > 8 - len(badge_locations):
+            for location in badge_locations:
+                if location.item:
+                    badge_items.append(location.item)
+                    location.item = None
+            continue
+        else:
+            break
+    else:
+        raise FillError(f"Failed to place badges for player {world.player}")
+    world.logic.guaranteed_hm_access = False
